# Third-party
import astropy.coordinates as coord
import astropy.units as u
import pytest
from astropy.time import Time

# Project
from ..barycenter import Barycenter


def test_init():
    b = Barycenter(
        origin=coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg, distance=100 * u.pc)
    )

    b = Barycenter(
        origin=coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg, distance=100 * u.pc),
        t0=59123.1213,
    )

    b = Barycenter(
        origin=coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg, distance=100 * u.pc),
        t0=Time(59123.1213, format="mjd"),
    )

    c = coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg, distance=100 * u.pc)
    with pytest.raises(ValueError):
        Barycenter(origin=c, radial_velocity=100 * u.km / u.s)

    with pytest.raises(ValueError):
        Barycenter()

    with pytest.raises(TypeError):
        Barycenter(origin=Time("J2000"))

    with pytest.raises(ValueError):
        Barycenter(
            origin=coord.SkyCoord(
                ra=[150, 100] * u.deg, dec=[-11, 10] * u.deg, distance=100 * u.pc
            )
        )

    with pytest.warns(UserWarning, match="Barycenter") as w:
        Barycenter(origin=coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg))



# Third-party
import astropy.units as u
import numpy as np
import pytest
from astropy.time import Time

# Project
from ..elements import KeplerElements, TwoBodyKeplerElements
from ..orbit import KeplerOrbit
from ..units import UnitSystem


def test_kepler():
    # Valid initialization:
    kw1 = dict(P=10 * u.day, e=0.5, omega=10 * u.deg, i=20 * u.deg, Omega=30 * u.deg)
    kw2 = dict(a=0.1 * u.au)
    kw3 = dict(M0=40 * u.deg, t0=Time("J2015.0"))
    kw4 = dict(units=UnitSystem(u.m, u.s, u.rad, u.kg))
    kw5 = dict(K=10 * u.km / u.s)

    for kw in [
        kw1,
        {**kw1, **kw2},
        {**kw1, **kw2, **kw3},
        {**kw1, **kw3},
        {**kw1, **kw4},
        {**kw1, **kw2, **kw4},
        {**kw1, **kw2, **kw3, **kw4},
        {**kw1, **kw5},
    ]:  # not meant to be exhaustive
        elems = KeplerElements(**kw)

        # Make sure this also works as arguments to KeplerOrbit
        KeplerOrbit(elems, elements_type="kepler")
        KeplerOrbit(**kw, elements_type="kepler")

    # Invalid initialization:
    for name in ["P", "omega"]:
        kw = kw1.copy()
        kw.pop(name)
        with pytest.raises(ValueError):
            KeplerElements(**kw)

    # check they get set to nan:
    kw = kw1.copy()
    kw.pop("Omega")
    elem = KeplerElements(**kw)
    assert np.isnan(elem.Omega)

    kw = kw1.copy()
    kw.pop("i")
    elem = KeplerElements(**kw)
    assert np.isnan(elem.i)

    # Make sure attributes exist and have correct units:
    elems = KeplerElements(
        P=10 * u.day,
        e=0.5,
        a=0.1 * u.au,
        omega=10 * u.deg,
        i=20 * u.deg,
        Omega=30 * u.deg,
        units=UnitSystem(u.m, u.s, u.rad, u.kg, u.m / u.s),
    )

    assert elems.P.unit == u.s
    assert elems.K.unit == u.m / u.s
    assert elems.m_f.unit == u.kg

    elems = KeplerElements(
        P=10 * u.day,
        e=0.5,
        K=10 * u.km / u.s,
        omega=10 * u.deg,
        i=20 * u.deg,
        Omega=30 * u.deg,
        units=UnitSystem(u.m, u.s, u.rad, u.kg, u.m / u.s),
    )

    assert elems.P.unit == u.s
    assert elems.K.unit == u.m / u.s
    assert elems.K == 10 * u.km / u.s
    assert elems.m_f.unit == u.kg

    # Expected failures:
    bad_kws = [("P", -10 * u.day), ("a", -10 * u.au), ("e", 1.5), ("i", -10 * u.deg)]
    for k, v in bad_kws:
        kw = kw1.copy()
        kw[k] = v
        with pytest.raises(ValueError):
            KeplerElements(**kw)

    with pytest.raises(ValueError):
        elems = KeplerElements(
            P=10 * u.day,
            e=0.5,
            K=10 * u.km / u.s,
            a=0.1 * u.au,
            omega=10 * u.deg,
            i=20 * u.deg,
            Omega=30 * u.deg,
        )

    # Check warning
    with pytest.warns(RuntimeWarning, match="Velocity"):
        KeplerElements(
            P=1 * u.day,
            a=5 * u.au,
            e=0.0,
            omega=10 * u.deg,
            i=20 * u.deg,
            Omega=30 * u.deg,
        )


def test_twobodykepler():
    TwoBodyKeplerElements(
        P=10 * u.day,
        e=0.5,
        m1=1 * u.Msun,
        m2=5 * u.Msun,
        omega=10 * u.deg,
        i=20 * u.deg,
        Omega=30 * u.deg,
    )
    TwoBodyKeplerElements(
        a=10 * u.au,
        e=0.5,
        m1=1 * u.Msun,
        m2=5 * u.Msun,
        omega=10 * u.deg,
        i=20 * u.deg,
        Omega=30 * u.deg,
    )

    # check attributes
    elems = TwoBodyKeplerElements(
        P=10 * u.day,
        e=0.5,
        m1=5 * u.Msun,
        m2=1 * u.Msun,
        omega=10 * u.deg,
        i=20 * u.deg,
        Omega=30 * u.deg,
    )

    with pytest.raises(AttributeError):
        elems.K

    with pytest.raises(AttributeError):
        elems.m_f

    elems.get_body("1")
    elems.get_body("2")

    elems.get_body(1)
    elems.get_body(2)

    with pytest.raises(ValueError):
        elems.get_body("derp")



from collections import defaultdict

import numpy as np
from setuptools import Extension, setup

exts = []

cfg = defaultdict(list)
cfg["include_dirs"].append(np.get_include())
cfg["include_dirs"].append("twobody/")
cfg["extra_compile_args"].append("--std=gnu99")
cfg["sources"].append("twobody/wrap.pyx")
cfg["sources"].append("twobody/src/twobody.c")
exts.append(Extension("twobody.wrap", **cfg))

setup(ext_modules=exts)



# Third-party
import astropy.coordinates as coord
import astropy.units as u
import numpy as np
import pytest
from astropy.tests.helper import quantity_allclose
from astropy.time import Time

try:
    import matplotlib.pyplot as plt

    HAS_MPL = True
except ImportError:
    HAS_MPL = False

# Package
from ..bary_trends import PolynomialRVTrend
from ..barycenter import Barycenter
from ..orbit import KeplerOrbit


def get_random_orbit_pars(rnd=None, barycenter=None):
    if rnd is None:
        rnd = np.random.RandomState()

    P = 10 ** rnd.uniform(0, 3) * u.day
    K = np.abs(rnd.normal(0, 100)) * u.km / u.s
    e = rnd.uniform(0, 1)
    i = rnd.uniform(0, 180) * u.deg
    n = 2 * np.pi / P
    a = K * np.sqrt(1 - e**2) / (n * np.sin(i))
    return dict(
        P=P,
        a=a,
        e=e,
        i=i,
        omega=rnd.uniform(0, 360) * u.deg,
        Omega=rnd.uniform(0, 360) * u.deg,
        M0=rnd.uniform(0, 360) * u.deg,
        barycenter=barycenter,
    )


def test_radial_velocity():
    # Compare the short-circuit RV to the full calculation

    rnd = np.random.RandomState(seed=42)
    n_iter = 32

    bc_kw = dict(ra=210.413231 * u.deg, dec=-61.3412 * u.deg, distance=101 * u.pc)

    # at rest w.r.t. sun
    bc1 = Barycenter(
        origin=coord.ICRS(
            **bc_kw,
            pm_ra_cosdec=0 * u.mas / u.yr,
            pm_dec=0.0 * u.mas / u.yr,
            radial_velocity=0 * u.km / u.s,
        ),
        t0=Time("J2018.0"),
    )

    bc2 = Barycenter(
        origin=coord.ICRS(
            **bc_kw,
            pm_ra_cosdec=1000 * u.mas / u.yr,
            pm_dec=1000.0 * u.mas / u.yr,
            radial_velocity=-260 * u.km / u.s,
        ),
        t0=Time("J2018.0"),
    )

    times = Time("J2018.0") + np.linspace(-100, 100, 128) * u.day

    # Even when barycenter is moving, rv1 and rv2 should be the same
    for n in range(n_iter):
        orb = KeplerOrbit(**get_random_orbit_pars(rnd, bc2))

        rv1 = orb.radial_velocity(times)
        _rp = orb.reference_plane(times)
        rv2 = -_rp.cartesian.differentials["s"].d_z + bc2.origin.radial_velocity

        assert quantity_allclose(rv1, rv2, rtol=0, atol=1e-3 * u.m / u.s)

    # All RV's should be equivalent
    for n in range(n_iter):
        orb = KeplerOrbit(**get_random_orbit_pars(rnd, bc1))

        rv1 = orb.radial_velocity(times)
        _rp = orb.reference_plane(times)
        rv2 = -_rp.cartesian.differentials["s"].d_z
        rv3 = orb.icrs(times).radial_velocity

        assert quantity_allclose(rv1, rv2, rtol=0, atol=1e-4 * u.m / u.s)
        # We expect true differences because of projection effects
        assert quantity_allclose(rv1, rv3, rtol=0, atol=1 * u.km / u.s)


def test_velocity_trend():
    rnd = np.random.RandomState(seed=42)

    coeffs = [0 * u.km / u.s, 0 * u.km / u.s / u.day, 1e-2 * u.km / u.s / u.day**2]
    trend = PolynomialRVTrend(coeffs, t0=Time("J2018.0"))
    pars = get_random_orbit_pars(rnd, trend)
    pars["a"] = 0 * u.au
    orbit = KeplerOrbit(**pars)

    t = Time("J2018.0") + np.linspace(-100, 100, 256) * u.day
    rv = orbit.radial_velocity(t)
    assert quantity_allclose(rv[0], rv[-1])
    assert quantity_allclose(rv[0], 100.0 * u.km / u.s)

    # OK THIS SUCKS (copy pasta)
    coeffs = [0 * u.km / u.s, 1 * u.km / u.s / u.day, 0 * u.km / u.s / u.day**2]
    trend = PolynomialRVTrend(coeffs, t0=Time("J2018.0"))
    pars = get_random_orbit_pars(rnd, trend)
    pars["a"] = 0 * u.au
    orbit = KeplerOrbit(**pars)

    t = Time("J2018.0") + np.linspace(-100, 100, 256) * u.day
    rv = orbit.radial_velocity(t)
    assert quantity_allclose(rv[0], -rv[-1])
    assert quantity_allclose(rv[0], -100.0 * u.km / u.s)


@pytest.mark.skipif(not HAS_MPL, reason="matplotlib not installed")
def test_plotting():
    orbit = KeplerOrbit(
        P=100 * u.day,
        a=1.0 * u.au,
        e=0.1,
        omega=0 * u.deg,
        i=41 * u.deg,
        Omega=0 * u.deg,
    )

    mjd = np.linspace(56823.123, 57293.2345, 1024)  # random MJD's
    t = Time(mjd, format="mjd", scale="utc")

    # plotting
    for _t in [mjd, t]:
        orbit.plot_rv(_t)

    orbit.plot_rv(t.jd, t_kwargs=dict(format="jd", scale="utc"))

    fig, ax = plt.subplots(1, 1)
    orbit.plot_rv(time=t, ax=ax, plot_kwargs=dict(color="#de2d26"))

    fig, ax = plt.subplots(1, 1)
    orbit.plot_rv(time=mjd, ax=ax)

    plt.close("all")


def test_nan():
    mjd = np.linspace(56123.123, 57293.2345, 1024)  # random MJD's
    t = Time(mjd, format="mjd", scale="utc")

    orb = KeplerOrbit(
        P=1.5 * u.year,
        e=0.67,
        M0=0 * u.deg,
        omega=17.14 * u.deg,
        i=np.nan * u.deg,
        Omega=np.nan * u.deg,
        a=np.nan * u.au,
    )
    rv = orb.radial_velocity(t)
    urv = orb.unscaled_radial_velocity(t)

    assert np.all(np.isnan(rv.value))
    assert np.all(np.isfinite(urv))

    orb = KeplerOrbit(P=1.5 * u.year, e=0.67, M0=0 * u.deg, omega=17.14 * u.deg)
    with pytest.raises(ValueError):
        rv = orb.radial_velocity(t)
    urv = orb.unscaled_radial_velocity(t)

    assert np.all(np.isnan(rv.value))
    assert np.all(np.isfinite(urv))



"""
A package for computing orbits and astronomical observables for binary stars,
exoplanets, and other gravitational two-body systems.
"""

from ._version import *
from .anomaly import *
from .bary_trends import *
from .barycenter import *
from .elements import *
from .orbit import *
from .reference_plane import *
from .transforms import *
from .units import *
